from trainyolo.utils.rle_cython import compute_rle
import numpy as np

def mask_to_rle(mask):
    counts = compute_rle(mask.flatten())
    size = [mask.shape[0], mask.shape[1]]
    return {"counts": counts, "size": size}    

def rle_to_mask(rle):
    (h, w), counts = rle['size'], rle['counts']

    mask = np.zeros(w*h, dtype=np.uint8)

    index = 0
    zeros = True
    for count in counts:
        if not zeros:
            mask[index : index + count] = 1
        index+=count
        zeros = not zeros

    mask = np.reshape(mask, [h, w])
    
    return mask


import os
import glob
from trainyolo.client import MLModel
import yaml
import re
import numpy as np
import cv2
from trainyolo.utils.ocr import read_f1_conf
from trainyolo.utils.rle import rle_to_mask, mask_to_rle
from trainyolo.utils.image_ops import scale_masks

def format_keypoints(boxes, keypoints, cls):
    annotations = []
    for box, kps, cl in zip(boxes, keypoints, cls):
        x1, y1, x2, y2 = box.tolist()
        w, h = x2 - x1, y2 - y1
        annotations.append({
            'bbox': [x1, y1, w, h],
            'area': w * h,
            'category_id': int(cl) + 1,
            'keypoints': [round(k) for kp in kps for k in kp]
        })
    return annotations

def format_boxes(boxes, cls):
    annotations = []
    for box, cl in zip(boxes, cls):
        x1, y1, x2, y2 = box.tolist()
        w, h = x2 - x1, y2 - y1
        annotations.append({
            'bbox': [x1, y1, w, h],
            'area': w * h,
            'category_id': int(cl) + 1
        })
    return annotations

def format_masks(masks, cls):
    annotations = []

    if masks:
        im_h, im_w = masks.orig_shape
        masks = masks.masks

        for mask, cl in zip(masks, cls):
            mask = scale_masks(mask[:,:,None], (im_h, im_w))[:,:,0]
            y, x = np.nonzero(mask)
            if len(y) > 0:
                x_min, x_max, y_min, y_max = int(np.min(x)), int(np.max(x)), int(np.min(y)), int(np.max(y))
                cropped_mask = mask[y_min:y_max+1, x_min:x_max+1]
                rle = mask_to_rle(cropped_mask>0)
                annotations.append({
                    'segmentation': rle,
                    'bbox': [x_min, y_min, x_max - x_min + 1, y_max - y_min + 1],
                    'category_id': int(cl) + 1
                })
    return annotations

def mask_to_polygons(mask, offset_x=0, offset_y=0, norm_x=1, norm_y=1):
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_TC89_KCOS)

    polygons = []
    for countour in contours:
        epsilon = 0.001 * cv2.arcLength(countour, True)
        polygon = cv2.approxPolyDP(countour, epsilon, True)
        polygon = ((polygon.squeeze() + np.array([offset_x, offset_y])) / np.array([norm_x, norm_y])).flatten().tolist()
        polygons.append(polygon)

    return polygons

# ref: ultralytics/JSON2YOLO
def min_index(arr1, arr2):
    """Find a pair of indexes with the shortest distance. 
    Args:
        arr1: (N, 2).
        arr2: (M, 2).
    Return:
        a pair of indexes(tuple).
    """
    dis = ((arr1[:, None, :] - arr2[None, :, :]) ** 2).sum(-1)
    return np.unravel_index(np.argmin(dis, axis=None), dis.shape)

# ref: ultralytics/JSON2YOLO
def merge_polygons(polygons):
    s = []
    polygons = [np.array(i).reshape(-1, 2) for i in polygons]
    idx_list = [[] for _ in range(len(polygons))]

    # record the indexes with min distance between each segment
    for i in range(1, len(polygons)):
        idx1, idx2 = min_index(polygons[i - 1], polygons[i])
        idx_list[i - 1].append(idx1)
        idx_list[i].append(idx2)

    # use two round to connect all the segments
    for k in range(2):
        # forward connection
        if k == 0:
            for i, idx in enumerate(idx_list):
                # middle segments have two indexes
                # reverse the index of middle segments
                if len(idx) == 2 and idx[0] > idx[1]:
                    idx = idx[::-1]
                    polygons[i] = polygons[i][::-1, :]

                polygons[i] = np.roll(polygons[i], -idx[0], axis=0)
                polygons[i] = np.concatenate([polygons[i], polygons[i][:1]])
                # deal with the first segment and the last one
                if i in [0, len(idx_list) - 1]:
                    s.append(polygons[i])
                else:
                    idx = [0, idx[1] - idx[0]]
                    s.append(polygons[i][idx[0]:idx[1] + 1])

        else:
            for i in range(len(idx_list) - 1, -1, -1):
                if i not in [0, len(idx_list) - 1]:
                    idx = idx_list[i]
                    nidx = abs(idx[1] - idx[0])
                    s.append(polygons[i][nidx:])
    return s


def annotations_to_yolo_polygons(annotations, im_w, im_h):
    output = []
    for ann in annotations:
        cl, bbox, rle = ann['category_id'], ann['bbox'], ann['segmentation']
        mask = rle_to_mask(rle)
        polygons = mask_to_polygons(mask, bbox[0], bbox[1], im_w, im_h)
        if len(polygons) > 1: # multi part polygon 
            merged_polygon = merge_polygons(polygons)
            merged_polygon = np.concatenate(merged_polygon, axis=0).flatten().tolist()
            output.append([cl-1, merged_polygon])
        elif len(polygons) > 0: # sometimes empty labels are saved
            output.append([cl-1, polygons[0]])
    return output

def upload_yolov8_run(project, mode='detect', run_location=None, run=None, weights='best.pt', conf=None, iou=0.45):
    run_location = run_location or './runs'
    run_location = os.path.join(run_location, mode)

    if mode == 'obb':
        # reset mode back to detect, as everything is the same
        mode = 'detect'

    # exp path
    if run is None:
        # get latest exp
        exp_paths = glob.glob(os.path.join(run_location, 'train*'))
        # order paths (natural ordering so a bit tricky)
        convert = lambda text: int(text) if text.isdigit() else text.lower()
        alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
        exp_path = sorted(exp_paths, key=alphanum_key)[-1]
    else:
        exp_path = os.path.join(run_location, run)

    # get result of best/last model
    csv_file = os.path.join(exp_path, 'results.csv')
    with open(csv_file, 'r') as f:
        lines = [[val.strip() for val in r.split(",")] for r in f.readlines()]
        headers, csv_list = lines[0], lines[1:]

        if mode == 'detect':
            results = [{'precision': float(item[4]), 'recall':float(item[5]), 'map50':float(item[6]), 'map':float(item[7])} for item in csv_list]
        elif mode == 'pose':
            results = [{
                'precision': float(item[6]), 
                'recall':float(item[7]), 
                'map50':float(item[8]), 
                'map':float(item[9]),
                'precision_pose': float(item[10]), 
                'recall_pose':float(item[11]), 
                'map50_pose':float(item[12]), 
                'map_pose':float(item[13]),
            } for item in csv_list]
        else: #segment
            results = [{
                'precision': float(item[5]), 
                'recall':float(item[6]), 
                'map50':float(item[7]), 
                'map':float(item[8]),
                'precision_mask': float(item[9]), 
                'recall_mask':float(item[10]), 
                'map50_mask':float(item[11]), 
                'map_mask':float(item[12]),
                } for item in csv_list]

    if weights == 'best.pt':
        # find best result
        best_result, best_fi = None, -1
        for result in results:
            if mode == 'detect':
                fi = 0.1*result['map50'] + 0.9*result['map']
            elif mode == 'pose':
                fi = 0.1*result['map50'] + 0.9*result['map'] + 0.1*result['map50_pose'] + 0.9*result['map_pose']
            else: #segment
                fi = 0.1*result['map50'] + 0.9*result['map'] + 0.1*result['map50_mask'] + 0.9*result['map_mask']
            if fi > best_fi:
                best_fi = fi
                best_result = result
        result = best_result
    else:
        # take last result
        result = results[-1]

    # get categories
    opt_file = os.path.join(exp_path, 'args.yaml')
    with open(opt_file, 'r') as f:
        opt = yaml.load(f, Loader=yaml.FullLoader)
    
    dataset_file = opt['data']
    with open(dataset_file, 'r') as f:
        dataset = yaml.load(f, Loader=yaml.FullLoader)
    categories = [{'id': id+1, 'name': name} for id, name in dataset['names'].items()]
    
    # read conf from f1 curve
    if conf is None:
        print('Reading best conf from F1_curve')
        if mode == 'detect':
            f1_curve = os.path.join(exp_path, 'F1_curve.png')
        elif mode == 'pose':
            f1_curve = os.path.join(exp_path, 'BoxF1_curve.png')
        else: 
            f1_curve = os.path.join(exp_path, 'BoxF1_curve.png')
        if os.path.exists(f1_curve):
            conf = read_f1_conf(project.client, f1_curve)
            if conf > 0:
                print(f'Using conf={conf}, which maximizes f1 score.')
                conf = conf
            else:
                print("Something went wrong while reading f1 curve, defaulting to conf=0.5")
                conf = 0.5
        else:
            print("Unable to find f1 curve, defaulting to conf=0.5")
            conf = 0.5

    # create model
    if project.model is None:
        model = MLModel.create(
            project.client,
            f'{project.name}',
            description='',
            type=project.annotation_type,
            public = project.public,
            project=project.uuid
        )
    else:
        model = project.model

    print(f'adding weights: {os.path.join(exp_path, "weights", weights)} to project ...')
    model.add_version(
        os.path.join(exp_path, 'weights', weights),
        categories=categories,
        architecture='yolov8' if mode == 'detect' else 'yolov8-seg' if mode == 'segment' else 'yolov8-pose',
        params={
            'model': opt['model'],
            'imgsz': opt['imgsz'],
            'conf': conf,
            'iou': iou
        },
        metrics={
            'map50': round(result['map50'], 3),
            'map': round(result['map'], 3),
            'precision': round(result['precision'], 3),
            'recall': round(result['recall'], 3)
        } if mode == 'detect' else {
            'map50': round(result['map50'], 3),
            'map': round(result['map'], 3),
            'precision': round(result['precision'], 3),
            'recall': round(result['recall'], 3),
            'map50_mask': round(result['map50_mask'], 3),
            'map_mask': round(result['map_mask'], 3),
            'precision_mask': round(result['precision_mask'], 3),
            'recall_mask': round(result['recall_mask'], 3)
        } if mode == 'segment' else {
            'map50': round(result['map50'], 3),
            'map': round(result['map'], 3),
            'precision': round(result['precision'], 3),
            'recall': round(result['recall'], 3),
            'map50_pose': round(result['map50_pose'], 3),
            'map_pose': round(result['map_pose'], 3),
            'precision_pose': round(result['precision_pose'], 3),
            'recall_pose': round(result['recall_pose'], 3) 
        }
    )


import requests
import os
from urllib.request import urlretrieve
from multiprocessing import Pool
from tqdm import tqdm
from itertools import repeat
import yaml
import time
import shutil
import json
from trainyolo.utils.bbox import get_corners_from_one_corner

class OCRResult:
    def __init__(self, client, data):
        self.client = client
        self.data = data

    def update(self):
        data = self.client.get(f'/functions/ocr/{self.uuid}/')
        self.data = data

    @property
    def uuid(self):
        return self.data['uuid']

    @property
    def result(self):
        return self.data['result']
    
    @property
    def processing_status(self):
        return self.data['processing_status']
    
    @property
    def processing_msg(self):
        return self.data['processing_msg']
    
    @classmethod
    def create(cls, client, image):

        asset = Asset.create(client, image)

        payload = {
            'image': asset.uuid
        }
        data = client.post(f'/functions/ocr/', payload)

        return cls(client, data)

    def get_result(self, retry=30):
        r = 0
        while(self.processing_status == 'PENDING'):
            self.update()
            time.sleep(1)
            r+=1
            if r > retry:
                break
        return self.processing_status, self.result

class Asset:

    def __init__(self, client, data):
        self.client = client
        self.data = data

    @property
    def uuid(self):
        return self.data['uuid']

    @property
    def filename(self):
        return self.data['filename']

    @property
    def asset(self):
        return self.data['asset']

    @classmethod
    def create(cls, client, file, type='OTHER'):
        # create asset
        filename = os.path.basename(file)
        payload = {
            'filename': filename,
            'type': type
        }
        data = client.post(f'/assets/', payload=payload)

        # upload file to s3
        url = data['presigned_post_fields']['url']
        fields = data['presigned_post_fields']['fields']
        
        with open(file, 'rb') as f:
            response = requests.post(url, data=fields, files={'file':f})
        
        if response.status_code != 204:
            raise Exception(f"Failed uploading to s3, status_code: {response.status_code}")

        # confirm upload
        data = client.put(f'/assets/{data["uuid"]}/confirm_upload/')

        return cls(client, data)

class MLModelVersion:

    def __init__(self, client, data):
        self.client = client
        self.data = data

    @property
    def version(self):
        return self.data['version']

    @property
    def categories(self):
        return self.data['categories']

    @property
    def architecture(self):
        return self.data['architecture']

    @property
    def weights(self):
        return self.data['weights']

    @property
    def params(self):
        return self.data['params']

    @property
    def metrics(self):
        return self.data['metrics']

    @property
    def created_at(self):
        return self.data['created_at']

    @classmethod
    def create(cls, client, model, weights, categories, architecture, params, metrics={}):

        # create asset
        asset = Asset.create(client, weights)

        # create version
        payload = {
            'categories': categories,
            'architecture': architecture,
            'weights': asset.uuid,
            'params': params,
            'metrics': metrics
        }
        data = client.post(f'/models/{model}/versions/', payload)

        return cls(client, data)

class MLModel:

    def __init__(self, client, data):
        self.client = client
        self.data = data

    @property
    def uuid(self):
        return self.data['uuid']

    @property
    def name(self):
        return self.data['name']

    @property
    def description(self):
        return self.data['description']

    @property
    def type(self):
        return self.data['type']

    @property
    def project(self):
        return Project.get_by_uuid(self.client, self.data['project'])

    @property
    def thumbnail(self):
        return self.data['thumbnail']

    @property
    def public(self):
        return self.data['public']

    @property
    def owner(self):
        return User(self.data['owner'])

    @property
    def latest_version(self):
        if self.data['latest_version']:
            return MLModelVersion(self.client, self.data['latest_version'])
        else:
            return None

    @property
    def created_at(self):
        return self.data['created_at']

    @classmethod
    def create(cls, client, name, description='', type='BBOX', public=False, project=None):
        payload = {
            'name': name,
            'description': description,
            'type': type,
            'public': public,
            'project': project
        }
        data = client.post('/models/', payload)

        return cls(client, data)

    @classmethod
    def get_or_create(cls, client, name, owner=None, **kwargs):
        try:
            model = cls.get_by_name(client, name, owner=owner)
        except:
            model = cls.create(client, name, **kwargs)

        return model

    @classmethod
    def get_by_uuid(cls, client, uuid):
        data = client.get(f'/models/{uuid}/')
        return cls(client, data)

    @classmethod
    def get_by_name(cls, client, name, owner=None):
        owner = owner if owner else client.owner.username
        data = client.get(f'/models/?name={name}&owner={owner}')
        
        if len(data) == 0:
            raise Exception(f'Did not found model "{name}" for owner "{owner}". If this is a shared or public model, please specify the owner')

        return cls(client, data[0])

    def add_version(self, weights, categories, architecture, params, metrics={}):
        return MLModelVersion.create(self.client, self.uuid, weights, categories, architecture, params, metrics=metrics)

class Video:

    def __init__(self, client, data):
        self.client = client
        self.data = data

    @property
    def uuid(self):
        return self.data['uuid']

    @property
    def name(self):
        return self.data['name']

    @property
    def video(self):
        return self.data['video']
    
    @property
    def created_at(self):
        return self.data['created_at']
    
    @classmethod
    def create(cls, client, project, video_file):
        # create asset
        asset = Asset.create(client, video_file)

        # create sample
        payload = {
            'name': asset.filename,
            'video': asset.uuid,
        }
        data = client.post(f'/projects/{project}/videos/', payload)

        return cls(client, data)
    

class Sample:

    def __init__(self, client, data):
        self.client  = client
        self.data = data

    @property
    def uuid(self):
        return self.data['uuid']

    @property
    def name(self):
        return self.data['name']

    @property
    def asset(self):
        return self.data['asset']

    @property
    def label(self):
        return self.data['label']

    @label.setter
    def label(self, annotations):
        payload = {
            'annotations': annotations
        }
        self.data['label'] = self.client.put(f'/samples/{self.uuid}/label/', payload=payload)

    @property
    def prediction(self):
        return self.data['prediction']
    
    @prediction.setter
    def prediction(self, annotations):
        payload = {
            'annotations': annotations
        }
        self.data['prediction'] = self.client.put(f'/samples/{self.uuid}/prediction/', payload=payload)

    @property
    def approved(self):
        return self.data['approved']
    
    @property
    def tags(self):
        return self.data['tags']

    @property
    def split(self):
        return self.data['split']

    @property
    def is_loading(self):
        return self.data['is_loading']

    @property
    def created_at(self):
        return self.data['created_at']
    
    @classmethod
    def create(cls, client, project, image_file, tags={}):
        # create asset
        asset = Asset.create(client, image_file, type='IMAGE')

        # create sample
        payload = {
            'name': asset.filename,
            'asset': asset.uuid,
            'tags': tags
        }
        data = client.post(f'/projects/{project}/samples/', payload)

        return cls(client, data)
    
    @classmethod
    def get_by_uuid(cls, client, uuid):
        data = client.get(f'/samples/{uuid}/')
        return cls(client, data)

    def pull_image(self, location='./'):
        asset_filename = self.asset['filename']
        asset_location = os.path.join(location, 'images', asset_filename)
        if not os.path.exists(asset_location):
            urlretrieve(self.asset['url'], asset_location)

    def pull_label(self, location='./', type='BBOX', format=None):
        if type in ['BBOX','KEYPOINT']:
            format = format or 'yolov8'
            if format in ['yolov5', 'yolov6', 'yolov8']:
                asset_filename = self.asset['filename']
                label_filename = os.path.splitext(asset_filename)[0] + '.txt'
                label_location = os.path.join(location, 'labels', label_filename)

                im_w, im_h = self.asset['metadata']['size'] 

                with open(label_location, 'w') as f:
                    for l in self.label['annotations']:
                        cl = l['category_id'] - 1
                        x_min, y_min, w, h = l['bbox']
                        x_max = x_min + w
                        y_max = y_min + h

                        x_min = min(max(0, x_min), im_w)
                        y_min = min(max(0, y_min), im_h)
                        x_max = min(max(0, x_max), im_w)
                        y_max = min(max(0, y_max), im_h)

                        xc, yc = (x_min + x_max) / 2.0, (y_min + y_max) / 2.0
                        w, h = abs(x_max - x_min), abs(y_max - y_min)
                        
                        xc, yc = xc / im_w, yc / im_h
                        w, h = w / im_w, h / im_h

                        f.write(f'{cl} {xc:.5f} {yc:.5f} {w:.5f} {h:.5f}')

                        if type == 'KEYPOINT':
                            keypoints = l['keypoints']
                            for kx, ky, kv in zip(keypoints[0::3], keypoints[1::3], keypoints[2::3]):
                                kx = min(max(0, kx), im_w) / im_w
                                ky = min(max(0, ky), im_h) / im_h
                                f.write(f' {kx:.5f} {ky:.5f} {kv}')

                        f.write('\n')
            else:
                raise Exception(f'Export format {format}" is not supported. Please check the documentation for the formats we support.')
        elif type=='BBOX_ROTATED':
                asset_filename = self.asset['filename']
                label_filename = os.path.splitext(asset_filename)[0] + '.txt'
                label_location = os.path.join(location, 'labels', label_filename)

                im_w, im_h = self.asset['metadata']['size'] 

                with open(label_location, 'w') as f:
                    for l in self.label['annotations']:
                        cl = l['category_id'] - 1
                        x_min, y_min, w, h, r = l['bbox']

                        # get rotated corners
                        corners = get_corners_from_one_corner(
                            x_min,
                            y_min,
                            w,
                            h,
                            r
                        )
                        
                        f.write(f'{cl}')
                        for corner in corners:
                            f.write(f' {max(0,min(corner[0]/im_w, 1)):.8f}')
                            f.write(f' {max(0, min(corner[1]/im_h, 1)):.8f}')
                        f.write('\n')
        
        elif type=='INSTANCE_SEGMENTATION':
            format = format or 'yolov8'
            if format in ['yolov8']:
                from trainyolo.utils.yolov8 import annotations_to_yolo_polygons
                
                asset_filename = self.asset['filename']
                label_filename = os.path.splitext(asset_filename)[0] + '.txt'
                label_location = os.path.join(location, 'labels', label_filename)

                im_w, im_h = self.asset['metadata']['size'] 

                with open(label_location, 'w') as f:
                    annotations = self.label['annotations']
                    yolo_polygons = annotations_to_yolo_polygons(annotations, im_w, im_h)

                    for cl, polygon in yolo_polygons:
                        f.write(f'{cl}')
                        for coord in polygon:
                            f.write(f' {coord:.5f}')
                        f.write('\n')
            else:
                raise Exception(f'Export format {format}" is not supported. Please check the documentation for the formats we support.')
        elif type=='SEGMENTATION':
            from trainyolo.utils.mask_utils import annotations_to_image

            asset_filename = self.asset['filename']
            label_filename = os.path.splitext(asset_filename)[0] + '.png'
            label_location = os.path.join(location, 'labels', label_filename)

            im_w, im_h = self.asset['metadata']['size'] 

            _, cls = annotations_to_image(self.label['annotations'], (im_w, im_h))
            cls.save(label_location)
        else:
            assert False, f'Type {type} is not supported.'

    def pull(self, location='./', type='BBOX', format=None):
        self.pull_image(location=location)
        self.pull_label(location=location, type=type, format=format)

    def delete_label(self):
        self.client.delete(f'/samples/{self.uuid}/label/')

    def delete(self):
        self.client.delete(f'/samples/{self.uuid}/')

class Project:
    
    def __init__(self, client, data):
        self.client = client
        self.data = data

    @property
    def uuid(self):
        return self.data['uuid']

    @property
    def name(self):
        return self.data['name']

    @property
    def description(self):
        return self.data['description']

    @property
    def thumbnail(self):
        return self.data['thumbnail']

    @property
    def categories(self):
        return self.data['categories']

    @property
    def annotation_type(self):
        return self.data['annotation_type']

    @property
    def num_samples(self):
        return self.data['num_samples']

    @property
    def model(self):
        if self.data['model']:
            return MLModel(self.client, self.data['model'])
        else:
            return None

    @property
    def prediction_model(self):
        if self.data['prediction_model']:
            return MLModel(self.client, self.data['prediction_model'])
        else:
            return None

    @property
    def samples(self):
        samples = self.get_samples(filter='')
        return samples

    @property
    def owner(self):
        return User(self.data['owner'])

    @property
    def collaborators(self):
        return [User(data) for data in self.data['collaborators']]

    @property
    def public(self):
        return self.data['public']

    @property
    def created_at(self):
        return self.data['created_at']

    @classmethod
    def create(cls, client, name, categories=['object'], description='', annotation_type='BBOX'):
        if annotation_type in ['BBOX', 'KEYPOINT']:
            category_dict = [{'id': id + 1, 'name': name} for id, name in enumerate(categories)]
        elif annotation_type == 'INSTANCE_SEGMENTATION':
            category_dict = [{'id': id + 1, 'name': name, 'has_instances': True} for id, name in enumerate(categories)]
        else:
            raise Exception(f'Annotation type "{annotation_type}" is not supported. Either use "BBOX" , "KEYPOINT" or "INSTANCE_SEGMENTATION".')

        payload = {
            'name': name,
            'description': description,
            'categories': category_dict,
            'annotation_type': annotation_type
        }
        data = client.post('/projects/', payload)

        return cls(client, data)

    @classmethod
    def get_or_create(cls, client, name, owner=None, *args, **kwargs):
        try:
            project = cls.get_by_name(client, name, owner=owner)
        except:
            project = cls.create(client, name, *args, **kwargs)

        return project

    @classmethod
    def get_by_uuid(cls, client, uuid):
        data = client.get(f'/projects/{uuid}/')
        return cls(client, data)

    @classmethod
    def get_by_name(cls, client, name, owner=None):
        owner = owner if owner else client.owner.username
        data = client.get(f'/projects/?name={name}&owner={owner}')
        
        if len(data) == 0:
            raise Exception(f'Did not found project "{name}" for owner "{owner}". If this is a shared or public project, please specify the owner')

        return cls(client, data[0])

    def pull(self, location='./', filter='LABELED', format=None, trainval_dirs=False):

        print('Downloading project')

        # make split
        self.make_split()

        # load samples
        samples = self.get_samples(filter=filter)

        # create output folders
        project_loc = os.path.join(location, self.name)
        image_loc = os.path.join(project_loc, 'images')
        label_loc = os.path.join(project_loc, 'labels')

        print('Downloading samples... (this may take a while depending on your dataset size)')
        if format == "coco":
            os.makedirs(image_loc, exist_ok=True)
            with Pool(8) as p:
                inputs = zip(samples, repeat(project_loc))
                r = p.starmap(Sample.pull_image, tqdm(inputs, total=len(samples)))
        else:        
            os.makedirs(image_loc, exist_ok=True)
            os.makedirs(label_loc, exist_ok=True)
            with Pool(8) as p:
                inputs = zip(samples, repeat(project_loc), repeat(self.annotation_type), repeat(format))
                r = p.starmap(Sample.pull, tqdm(inputs, total=len(samples)))

        if format == "coco":           
            train_images, val_images = [], []
            train_annotations, val_annotations = [], []

            ann_id = 0
            for image_id, sample in enumerate(samples):
                image = {
                    'id': image_id,
                    'file_name': sample.name,
                    'width': sample.asset['metadata']['size'][0],
                    'height': sample.asset['metadata']['size'][1]
                }
                if sample.split == 'TRAIN':
                    train_images.append(image)
                else:
                    val_images.append(image)

                if sample.label is not None:
                    for item in sample.label['annotations']:
                        ann = {
                            'id': ann_id,
                            'image_id': image_id,
                            'bbox': item['bbox'],
                            'area': item['bbox'][2] * item['bbox'][3],
                            'iscrowd': 0,
                            'category_id': item['category_id']
                        }
                        if sample.split == 'TRAIN':
                            train_annotations.append(ann)
                        else:
                            val_annotations.append(ann)
                        
                        ann_id += 1

            with open(os.path.join(project_loc, 'train.json'), 'w') as f:
                coco = {
                    'images': train_images,
                    'categories': self.categories,
                    'annotations': train_annotations
                }
                json.dump(coco, f)
            
            with open(os.path.join(project_loc, 'val.json'), 'w') as f:
                coco = {
                    'images': val_images,
                    'categories': self.categories,
                    'annotations': val_annotations
                }
                json.dump(coco, f)
            
            return

        # create train-val txt file
        with open(os.path.join(project_loc, 'train.txt'), 'w') as f_train, open(os.path.join(project_loc, 'val.txt'), 'w') as f_val:
            for s in samples:
                if s.split == 'TRAIN':
                    f_train.write(f'./images/{s.asset["filename"]}\n')
                if s.split == 'VAL':
                    f_val.write(f'./images/{s.asset["filename"]}\n')

        # create dataset yaml file
        if format == "yolov6":
            with open(os.path.join(project_loc, 'dataset.yaml'), 'w') as f:
                yaml_content = {
                    'train': os.path.join(os.path.abspath(project_loc), 'images', 'train'),
                    'val': os.path.join(os.path.abspath(project_loc), 'images', 'val'),
                    'is_coco': False,
                    'nc': len(self.categories),
                    'names': [cat['name'] for cat in self.categories]
                }
                yaml.dump(yaml_content, f, default_flow_style=False)
        else:
            with open(os.path.join(project_loc, 'dataset.yaml'), 'w') as f:
                yaml_content = {
                    'path': os.path.abspath(project_loc),
                    'train': 'train.txt',
                    'val': 'val.txt',
                    'names': {cat['id']-1:cat['name'] for cat in self.categories}
                }
                if self.annotation_type == 'KEYPOINT':
                    yaml_content['kpt_shape'] = [len(self.categories[0]['keypoints']),3]
                yaml.dump(yaml_content, f, default_flow_style=False)

        # create train/val folder with symlinks
        if format == "yolov6":

            # remove folders as train/val spit might have changed
            [shutil.rmtree(os.path.join(project_loc, f, split)) for f in ['images', 'labels'] for split in ['train', 'val'] if os.path.exists(os.path.join(project_loc, f, split))]

            # create folders
            [os.makedirs(os.path.join(project_loc, f, split), exist_ok=True) for f in ['images', 'labels'] for split in ['train', 'val'] ]

            # create symlinks
            for s in samples:
                txt_filename = s.asset["filename"].rsplit('.',1)[0] + '.txt'
                os.symlink(f'../{s.asset["filename"]}', f'{os.path.abspath(project_loc)}/images/{"train" if s.split == "TRAIN" else "val"}/{s.asset["filename"]}')
                os.symlink(f'../{txt_filename}', f'{os.path.abspath(project_loc)}/labels/{"train" if s.split == "TRAIN" else "val"}/{txt_filename}')

        # create seperate train/val folder with symlinks
        if trainval_dirs:
            
            # remove folders as train/val spit might have changed
            [shutil.rmtree(os.path.join(project_loc, split, f)) for split in ['train', 'val'] for f in ['images', 'labels'] if os.path.exists(os.path.join(project_loc, split, f))]

            # create folders
            [os.makedirs(os.path.join(project_loc, split, f), exist_ok=True) for split in ['train', 'val'] for f in ['images', 'labels']]

            # create symlinks
            for s in samples:
                txt_filename = s.asset["filename"].rsplit('.',1)[0] + '.txt'
                os.symlink(f'../../images/{s.asset["filename"]}', f'{os.path.abspath(project_loc)}/{"train" if s.split == "TRAIN" else "val"}/images/{s.asset["filename"]}')
                os.symlink(f'../../labels/{txt_filename}', f'{os.path.abspath(project_loc)}/{"train" if s.split == "TRAIN" else "val"}/labels/{txt_filename}')
        
        return project_loc


    def push(self, image_list):
        with Pool(8) as p:
            inputs = zip(repeat(self), image_list)
            r = p.starmap(Project.add_sample, tqdm(inputs, total=len(image_list)))

    def get_samples(self, filter='LABELED'):

        if filter.lower() == 'labeled':
            query = 'labeled=True'
        elif filter.lower() == 'unlabeled':
            query = 'labeled=False'
        else:
            query = ''

        sample_list = self.client.get(f'/projects/{self.uuid}/sample_list/?{query}')
        
        print('Loading samples...')
        with Pool(8) as p:
            samples = p.starmap(Sample.get_by_uuid, tqdm(zip(repeat(self.client), sample_list), total=len(sample_list)))
        
        return samples

    def add_sample(self, image_file, tags={}):
        return Sample.create(self.client, self.uuid, image_file, tags=tags)

    def make_split(self):
        return self.client.put(f'/projects/{self.uuid}/make_split/')

    def delete(self):
        self.client.delete(f'/projects/{self.uuid}/')

class User:
    def __init__(self, data):
        self.data = data

    @property
    def username(self):
        return self.data['username']

    @property
    def email(self):
        return self.data['email']

class Client:

    def __init__(self, apikey, url="https://api.trainyolo.com"):
        self.url = url
        self.apikey = apikey

    @property
    def owner(self):
        data = self.get('/auth/users/me')
        return User(data)
    
    def get(self, endpoint):

        response = requests.get(self.url + endpoint, headers=self._get_headers())
        
        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(f"Request failed, status_code: {response.status_code}")

    def post(self, endpoint, payload={}, files=None):

        response = requests.post(self.url + endpoint, json=payload, files=files, headers=self._get_headers())

        if response.status_code == 201:
            return response.json()
        else:
            raise Exception(f"Request failed, status_code: {response.status_code} - {response.text}")

    def patch(self, endpoint, payload={}):
        
        response = requests.patch(self.url + endpoint, json=payload, headers = self._get_headers())

        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(f"Request failed, status_code: {response.status_code} - {response.text}")

    def put(self, endpoint, payload={}):

        response = requests.put(self.url + endpoint, json=payload, headers = self._get_headers())

        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(f"Request failed, status_code: {response.status_code} - {response.text}")

    def delete(self, endpoint):
        response = requests.delete(self.url + endpoint, headers=self._get_headers())
        
        if response.status_code != 204:
            raise Exception(f"Request failed, status_code: {response.status_code}")

    def _get_headers(self):

        # set content type & authorization token
        headers = {
            'Authorization': f"APIKey {self.apikey}"
        }

        return headers




#cython: language_level=3, boundscheck=False, wraparound=False, nonecheck=False
import numpy as np
cimport numpy as np

def compute_rle(np.ndarray[np.uint8_t, ndim=1] mask):

    counts = []
    cdef int prev_val = 0
    cdef int cnt = 0

    cdef Py_ssize_t i
    for i in range(mask.shape[0]):
        if mask[i] == prev_val:
            cnt +=1
        else:
            counts.append(cnt)
            prev_val = mask[i]
            cnt = 1
    counts.append(cnt)

    return counts


import os
import glob
from trainyolo.client import MLModel
import yaml
import re
from trainyolo.utils.ocr import read_f1_conf

def format_boxes(boxes):
    annotations = []
    for box in boxes:
        x1, y1, x2, y2, _, cl = box.tolist()
        w, h = x2 - x1, y2 - y1
        annotations.append({
            'bbox': [x1, y1, w, h],
            'area': w * h,
            'category_id': int(cl) + 1
        })
    return annotations

def upload_yolov5_run(project, run_location=None, run=None, weights='best.pt', conf=None, iou=0.45):
    run_location = run_location or './runs'
    run_location = os.path.join(run_location, 'train')
    # exp path
    if run is None:
        # get latest exp
        exp_paths = glob.glob(os.path.join(run_location, '*'))
        # order paths (natural ordering so a bit tricky)
        convert = lambda text: int(text) if text.isdigit() else text.lower()
        alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
        exp_path = sorted(exp_paths, key=alphanum_key)[-1]
    else:
        exp_path = os.path.join(run_location, run)

    # get result of best/last model
    csv_file = os.path.join(exp_path, 'results.csv')
    with open(csv_file, 'r') as f:
        lines = [[val.strip() for val in r.split(",")] for r in f.readlines()]
        headers, csv_list = lines[0], lines[1:]
        results = [{'precision': float(item[4]), 'recall':float(item[5]), 'map50':float(item[6]), 'map':float(item[7])} for item in csv_list]

    if weights == 'best.pt':
        # find best result
        best_result, best_fi = None, -1
        for result in results:
            fi = 0.1*result['map50'] + 0.9*result['map']
            if fi > best_fi:
                best_fi = fi
                best_result = result
        result = best_result
    else:
        # take last result
        result = results[-1]

    # get categories
    opt_file = os.path.join(exp_path, 'opt.yaml')
    with open(opt_file, 'r') as f:
        opt = yaml.load(f, Loader=yaml.FullLoader)
    
    dataset_file = opt['data']
    with open(dataset_file, 'r') as f:
        dataset = yaml.load(f, Loader=yaml.FullLoader)
    categories = [{'id': id+1, 'name': name} for id, name in dataset['names'].items()]

    # read conf from f1 curve
    if conf is None:
        print('Reading best conf from F1_curve')
        f1_curve = os.path.join(exp_path, 'F1_curve.png')
        if os.path.exists(f1_curve):
            conf = read_f1_conf(project.client, f1_curve)
            if conf > 0:
                print(f'Using conf={conf}, which maximizes f1 score.')
                conf = conf
            else:
                print("Something went wrong while reading f1 curve, defaulting to conf=0.5")
                conf = 0.5
        else:
            print("Unable to find f1 curve, defaulting to conf=0.5")
            conf = 0.5
    
    # create model
    if project.model is None:
        model = MLModel.create(
            project.client,
            f'{project.name}',
            description='',
            type='BBOX',
            public = project.public,
            project=project.uuid
        )
    else:
        model = project.model

    print('adding version to project ...')
    model.add_version(
        os.path.join(exp_path, 'weights', weights),
        categories=categories,
        architecture='yolov5',
        params={
            'model': opt['weights'],
            'imgsz': opt['imgsz'],
            'conf': conf,
            'iou': iou
        },
        metrics={
            'map50': round(result['map50'], 3),
            'map': round(result['map'], 3),
            'precision': round(result['precision'], 3),
            'recall': round(result['recall'], 3)
        }
    )


from trainyolo.client import MLModel
import yaml
import os
from datetime import datetime

def upload_latest_run(project, run_location=None):
    run_location = run_location or './output'

    # get latest run
    runs = os.listdir(os.path.join(run_location, 'train'))
    runs.sort(key=lambda date: datetime.strptime(date, "%m-%d-%Y-%H:%M:%S"), reverse=True)

    run_path = os.path.join(run_location, 'train', runs[0])    

    # read config 
    with open(os.path.join(run_path, 'config.yaml'), 'r') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)

    # read results
    with open(os.path.join(run_path, 'best_miou_model.csv'), 'r') as f:
        lines = [[val.strip() for val in r.split(",")] for r in f.readlines()]
        header, values = lines[0], lines[1]

    # create model
    if project.model is None:
        model = MLModel.create(
            project.client,
            f'{project.name}',
            description='',
            type=project.annotation_type,
            public = project.public,
            project=project.uuid
        )
    else:
        model = project.model

    print(f'adding weights: {os.path.join(run_path, "best_miou_model.pt")} to project ...')
    model.add_version(
        os.path.join(run_path, "best_miou_model.pt"),
        categories=project.categories,
        architecture='trainyolo-seg',
        params={
            'model': config['model'],
            'encoder': config['model_encoder'],
            'min_size': config['img_size'],
            'pad': 32,
            'normalize': True,
        },
        metrics={
        'miou': round(float(values[0]), 5),
        'class_iou': {cl:round(float(v),5) for cl, v in zip(header[1:], values[1:])}
        }
    )


import numpy as np

def rotate_around_point(point, angle, origin):
    # Convert angle to radians
    angle_rad = np.radians(angle)

    # Translation to the origin
    translated = point - origin

    # Rotation matrix
    rotation_matrix = np.array([[np.cos(angle_rad), -np.sin(angle_rad)],
                                [np.sin(angle_rad), np.cos(angle_rad)]])

    # Perform rotation and translate back
    rotated = np.dot(rotation_matrix, translated) + origin
    return rotated

def get_corners_from_one_corner(x0, y0, w, h, rotation):
    # Calculate other three corners relative to x0, y0
    corners = np.array([[x0, y0], [x0 + w, y0], [x0, y0 + h], [x0 + w, y0 + h]])

    # Rotate the corners around x0, y0
    rotated_corners = []
    for corner in corners:
        rotated = rotate_around_point(corner, rotation, np.array([x0, y0]))
        rotated_corners.append(rotated)

    return rotated_corners


from trainyolo.client import MLModel

def upload_yolonas_run(project, checkpoint_path, model_name, imgsz=640, conf=0.5, iou=0.45, metrics={}):

    # create model
    if project.model is None:
        model = MLModel.create(
            project.client,
            f'{project.name}',
            description='',
            type=project.annotation_type,
            public = project.public,
            project=project.uuid
        )
    else:
        model = project.model

    # add version
    print(f'adding checkpoint: {checkpoint_path} to project ...')
    model.add_version(
        checkpoint_path,
        categories=project.categories,
        architecture='yolonas',
        params={
            'model': model_name,
            'imgsz': imgsz,
            'conf': conf,
            'iou': iou
        },
        metrics=metrics
    )



from trainyolo.client import OCRResult
import re

def read_f1_conf(client, f1_curve):
    ocr_result = OCRResult.create(client, f1_curve)
    status, result = ocr_result.get_result()
    if status == 'SUCCEEDED':
        pattern = re.compile(r'all\s+classes\s+(\d+\.\d+)\s+at\s+(\d+\.\d+)')
        match = pattern.search(result)
        if match:
            score, conf = match.groups()
            conf = float(conf)
            if conf < 1 and conf > 0:
                return conf
            else:
                return -1
    else:
        return -1


import numpy as np
from trainyolo.utils.rle import rle_to_mask
from PIL import Image
from trainyolo.utils.colormap import get_colormap

def annotations_to_image(annotations, size):

    w, h = size
    inst = np.zeros((h, w), dtype=np.uint16)
    cls = np.zeros((h, w), dtype=np.uint8)

    for i, ann in enumerate(annotations):
        mask = rle_to_mask(ann['segmentation'])
        
        x, y, w, h = ann['bbox']

        # hack for overlapping masks
        inst_region = inst[y:y+h,x:x+w]
        mask = mask * (inst_region == 0)

        inst[y:y+h,x:x+w] += mask * (i+1) 
        cls[y:y+h,x:x+w] += mask * ann["category_id"]

    cls = Image.fromarray(cls, mode='P')
    cls.putpalette(get_colormap())

    inst = Image.fromarray(inst, mode='I;16')

    return inst, cls


import sys
from argparse import ArgumentParser
import argcomplete

class Options:

    def __init__(self):

        self.parser = ArgumentParser(
            description="Cli for trainYOLO."
        )

        subparsers = self.parser.add_subparsers(dest="command")

        # authenticate
        subparsers.add_parser("authenticate", help="Store apikey for authentication.")

        # project
        project = subparsers.add_parser("project", help="project related functions")
        project_action = project.add_subparsers(dest="action")

        ## project create
        project_create = project_action.add_parser("create", help="Create a new project")
        project_create.add_argument(
            "name",
            type=str,
            help="Name of project"
        )
        project_create.add_argument(
            "--type",
            type=str,
            default="BBOX",
            help="project type: BBOX or INSTANCE_SEGMENTATION"
        )
        project_create.add_argument(
            "--categories",
            type=str,
            default="object",
            help="list of categories, use comma between categories, eg: cat_1,cat_2"
        )

        ## project pull
        project_pull = project_action.add_parser("pull", help="Pull samples from project")
        project_pull.add_argument(
            "project",
            type=str,
            help="Name of project"
        )
        project_pull.add_argument(
            "--path",
            type=str,
            default='./',
            help="Path to image folder"
        )
        project_pull.add_argument(
            "--format",
            type=str,
            help="Output format"
        )
        project_pull.add_argument(
            "--owner",
            type=str,
            help="project owner"
        )

        ## project push
        project_push = project_action.add_parser("push", help="Push images to project")
        project_push.add_argument(
            "project",
            type=str,
            help="Name of project"
        )
        project_push.add_argument(
            "path",
            type=str,
            help="Path to image folder"
        )

        ## project add_model
        project_add_model = project_action.add_parser("add_model", help="add model to project")
        project_add_model.add_argument(
            "project",
            type=str,
            help="Name of project"
        )

        project_add_model.add_argument(
            "--type",
            type=str,
            default='yolov5',
            help="Model type, eg yolov5/yolov8/yolov8-seg/yolov8-pose"
        )

        project_add_model.add_argument(
            "--run-location",
            type=str,
            help="Run location, typically ./runs"
        )
        project_add_model.add_argument(
            "--run",
            type=str,
            help="Run, eg exp (optional, if omitted will use latest run)"
        )
        project_add_model.add_argument(
            "--conf",
            type=float,
            default=None,
            help="conf value, if ommited will use the one that maximizes f1 score"
        )
        project_add_model.add_argument(
            "--iou",
            type=float,
            default=.45,
            help="Nms threshold value"
        )

        # version
        subparsers.add_parser("version", help="Print current version number")

        argcomplete.autocomplete(self.parser)

    def parse_args(self):
        args = self.parser.parse_args()

        if not args.command:
            self.parser.print_help()
            sys.exit()

        return args, self.parser


import os
from pathlib import Path
from trainyolo.client import Client, Project
from trainyolo.utils.yolov5 import upload_yolov5_run
from trainyolo.utils.yolov8 import upload_yolov8_run
import glob
import sys

def _save_apikey(apikey):
    config_path = Path().home() / '.trainyolo' / 'credentials'
    config_path.parent.mkdir(exist_ok=True)

    with open(config_path, 'w') as f:
        f.write(f"apikey={apikey}\n")

def _load_apikey():
    config_path = Path().home() / '.trainyolo' / 'credentials'
    
    try:
        with open(config_path, 'r') as f:
            apikey = f.readline().split('=',1)[1].strip()
            return apikey
    except:
        print('No API Key detected. First authenticate.')
        sys.exit(1)

def _get_client():
    apikey = _load_apikey()
    client = Client(apikey)
    return client

def authenticate(apikey):
    try:
        client = Client(apikey)
        _save_apikey(apikey)
    except:
        print('API Key is invalid.')
        sys.exit(1)

def create_project(name, annotation_type, categories):
    client = _get_client()
    try:
        categories = [cat.strip() for cat in categories.split(',')]
        Project.create(client, name, categories=categories, annotation_type=annotation_type)
    except Exception as e:
        print(e)
        sys.exit(1)

def pull_project(name, path, format, owner):
    client = _get_client() 
    try:   
        project = Project.get_by_name(client, name, owner=owner)
        project.pull(location=path, format=format)
    except Exception as e:
        print(e)
        sys.exit(1)

def push_to_project(name, path):
    client = _get_client()
    try:
        project = Project.get_by_name(client, name)
        
        image_files = []
        image_types = (
            '*.png', '*.PNG', 
            '*.jpg','*.JPG', 
            '*.jpeg', '*.JPEG', 
            '*.tiff', '*.TIFF', 
            '*.tif', '*.TIF'
        )
        for t in image_types:
            image_files.extend(glob.glob(os.path.join(path, t)))
        image_files.sort()

        project.push(image_files)

    except Exception as e:
        print(e)
        sys.exit(1)

def add_model(name, type, run_location, run, conf, iou):
    client = _get_client()
    try:
        project = Project.get_by_name(client, name)
        if type == 'yolov5':
            upload_yolov5_run(project, run_location=run_location, run=run, conf=conf, iou=iou)
        elif type == 'yolov8':
            upload_yolov8_run(project, mode='detect', run_location=run_location, run=run, conf=conf, iou=iou)
        elif type == 'yolov8-obb':
            upload_yolov8_run(project, mode='obb', run_location=run_location, run=run, conf=conf, iou=iou)
        elif type == 'yolov8-seg':
            upload_yolov8_run(project, mode='segment', run_location=run_location, run=run, conf=conf, iou=iou)
        elif type == 'yolov8-pose':
            upload_yolov8_run(project, mode='pose', run_location=run_location, run=run, conf=conf, iou=iou)
        else:
            print(f'Type "{type}" is curently not supported. Options are: yolov5|yolov8|yolov8-seg|yolov8-pose')
            sys.exit(1)
    except Exception as e:
        print(e)
        sys.exit(1)


import cv2

# taken from https://github.com/ultralytics/ultralytics/blob/main/ultralytics/yolo/utils/ops.py
def scale_masks(masks, im0_shape, ratio_pad=None):
    """
    Takes a mask, and resizes it to the original image size
    Args:
      masks (torch.Tensor): resized and padded masks/images, [h, w, num]/[h, w, 3].
      im0_shape (tuple): the original image shape
      ratio_pad (tuple): the ratio of the padding to the original image.
    Returns:
      masks (torch.Tensor): The masks that are being returned.
    """
    # Rescale coordinates (xyxy) from im1_shape to im0_shape
    im1_shape = masks.shape
    if im1_shape[:2] == im0_shape[:2]:
        return masks
    if ratio_pad is None:  # calculate from im0_shape
        gain = min(im1_shape[0] / im0_shape[0], im1_shape[1] / im0_shape[1])  # gain  = old / new
        pad = (im1_shape[1] - im0_shape[1] * gain) / 2, (im1_shape[0] - im0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]
    top, left = int(pad[1]), int(pad[0])  # y, x
    bottom, right = int(im1_shape[0] - pad[1]), int(im1_shape[1] - pad[0])

    if len(masks.shape) < 2:
        raise ValueError(f'"len of masks shape" should be 2 or 3, but got {len(masks.shape)}')
    masks = masks[top:bottom, left:right]
    # masks = masks.permute(2, 0, 1).contiguous()
    # masks = F.interpolate(masks[None], im0_shape[:2], mode='bilinear', align_corners=False)[0]
    # masks = masks.permute(1, 2, 0).contiguous()
    masks = cv2.resize(masks, (im0_shape[1], im0_shape[0]), interpolation=cv2.INTER_NEAREST)
    if len(masks.shape) == 2:
        masks = masks[:, :, None]

    return masks


from trainyolo.cli_options import Options
from trainyolo import cli_functions
import getpass
from trainyolo import __version__


def main():
    
    args, parser = Options().parse_args()
    exec_command(args, parser)

def exec_command(args, parser):

    if args.command == "authenticate":
        apikey = getpass.getpass(prompt="API key: ", stream=None)
        apikey = apikey.strip()
        if apikey == "":
            print("No API key entered. Consult the trainyolo documentation on how to get your API key.")
            return
        else:
            cli_functions.authenticate(apikey)

    if args.command == "version":
        print(__version__)

    if args.command == "project": 
        if args.action == "create":
            cli_functions.create_project(args.name, args.type, args.categories)
        elif args.action == "pull":
            cli_functions.pull_project(args.project, args.path, args.format, args.owner)
        elif args.action == "push":
            cli_functions.push_to_project(args.project, args.path)
        elif args.action == "add_model":
            cli_functions.add_model(args.project, args.type, args.run_location, args.run, args.conf, args.iou)

        
if __name__ == "__main__":
    main()